;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Maria Vincit
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; program : Maria Vincit BP -> 3d rotozoomer. Textured
;;           with bilinear filtering, tetrahedron,
;;           illuminated by three unlinear lights.
;; author  : Maciej Guba aka macgub
;; web     : www.macgub.vxm.pl
;; email   : macgub3@wp.pl
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Credits:
;;   Big thanks to Jan Pawel II, Tomasz Grysztar, Mikolaj
;;   Feliks, Maria Vincit publishing house, Ica, Chem,
;;   lostcauz, tthsqe and all from board.flatassembler.net.
;;   It's 32 bit sse3 app.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

format PE GUI 4.0
entry start

include '/include/win32a.inc'

TEX_X = 128    ; texture size
TEX_Y = 128
TEXTURE_SIZE EQU (TEX_X * TEX_Y)-1
TEX_SHIFT = 7
ROUND equ 8
WM_FINISH equ WM_USER+100h

section '.text' code readable executable
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; start and message loop
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
start:

    invoke GetModuleHandle,0
    mov [wc.hInstance],eax
    invoke LoadIcon,0,IDI_APPLICATION
    mov [wc.hIcon],eax
    invoke LoadCursor,0,IDC_ARROW
    mov [wc.hCursor],eax

    invoke  GetSystemMetrics,SM_CXSCREEN
    sub     eax,10
    and     eax,0xfffffffe
    push    eax
    mov     [xres_var],ax
    shr     eax,1
    mov     [xxadd],ax
    invoke  GetSystemMetrics,SM_CYSCREEN
    sub     eax,10
    and     eax,0xfffffffe
    push    eax
    mov     [yres_var],ax
    mov     ecx,eax
    shr     ecx,1
    mov     [yyadd],cx

    pop eax
    pop ebx

    mov [bmi.biSize],sizeof.BITMAPINFOHEADER
    mov [bmi.biWidth],ebx
    neg eax
    mov [bmi.biHeight],eax
    mov [bmi.biPlanes],1
    mov [bmi.biBitCount],32
    mov [bmi.biCompression],BI_RGB

    mov [bmi.biSizeImage],0
    mov [bmi.biXPelsPerMeter],0
    mov [bmi.biYPelsPerMeter],0
    mov [bmi.biClrUsed],0
    mov [bmi.biClrImportant],0

    mov [rect.left],0
    mov [rect.top],0
    mov [rect.right],ebx
    neg eax
    mov [rect.bottom],eax
    invoke AdjustWindowRect,rect,WS_POPUP or WS_SYSMENU or WS_CAPTION,0 

    invoke RegisterClass,wc

    mov ecx,[rect.bottom]
    mov eax,[rect.right]
    sub ecx,[rect.top]
    sub eax,[rect.left] 

    invoke CreateWindowEx,0,_class,_title,WS_OVERLAPPEDWINDOW\
           xor WS_THICKFRAME xor WS_MAXIMIZEBOX,0,0,eax,ecx,NULL,NULL,0,0
    mov [hwnd],eax
    invoke GetDC,[hwnd]
    mov [hdc],eax
    invoke ShowWindow,[hwnd],SW_SHOW


     movzx   eax,word[xres_var]
     movzx   ebx,[yres_var]
     mul     ebx
     shl     eax,2
     mov     ebx,eax
     invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     or  eax,eax
     jnz @f
     int3
   @@:
     mov     [screen_ptr],eax

     invoke  VirtualAlloc,NULL,ebx,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     or  eax,eax
     jnz @f
     int3
   @@:
     mov     [Zbuffer_ptr],eax

     movzx    eax,[yres_var] ;YRES / 4
     shr      eax,2
     cvtsi2ss xmm0,eax
     movss  [scale],xmm0

;    mov esi,mv
;    mov edi,texmap
;    mov ecx,TEX_Y*TEX_X
;  @@:
;    mov eax,[esi]
;    mov [edi],eax
;    add edi,4
;    add esi,3
;    sub ecx,1
;    jnz @b


   mov  esi,mv16
   mov  edi,texmap
   mov  ecx,TEX_X*TEX_Y/2
   add  esi,ecx
  @@:
   movzx  eax,byte[esi]
   and    eax,0x0000000f
   lea    eax,[eax*3]
   add    eax,palette
   mov    eax,[eax]
   and    eax,0x00ffffff
   mov    [edi],eax

   movzx  eax,byte[esi]
   shr    eax,4
   lea    eax,[eax*3]
   add    eax,palette
   mov    eax,[eax]
   and    eax,0x00ffffff
   mov    [edi+4],eax

   sub   esi,1
   add   edi,8
   sub   ecx,1
   jnz   @b

   mov  edi,lights_aligned
   mov  ecx,3
   call norm_aligned_lv


    pushad
    push   dword points_r
    movzx  ecx,[points_count_var]
    push   ecx
    call   normalize_object1
    popad

    call   init_triangles_normals2
    call   init_point_normals

msg_loop:

    stdcall g_draw
    inc  [ccounter]
    and  [ccounter],0x000001ff
    movzx  eax,word[xres_var]
    movzx  ebx,[yres_var]
    invoke StretchDIBits,[hdc],0,0,eax,ebx,0,0,eax,ebx,[screen_ptr],bmi,0,SRCCOPY

    cmp    [speed_flag],1
    je     @f
    invoke Sleep,1
  @@:

    invoke PeekMessage,msg,0,0,0,PM_NOREMOVE
    or eax,eax
    je msg_loop

    invoke GetMessage,msg,NULL,0,0
    or eax,eax
    je end_loop

    invoke TranslateMessage,msg
    invoke DispatchMessage,msg
    jmp msg_loop

error:
    invoke MessageBox,NULL,_error,NULL,MB_ICONERROR+MB_OK

end_loop:
    invoke ReleaseDC,[hwnd],[hdc]
    invoke DestroyWindow,[hwnd]
    invoke ExitProcess,[msg.wParam]
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; WindowProc
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
proc WindowProc uses ebx esi edi, hwnd,wmsg,wparam,lparam
    cmp [wmsg],WM_KEYDOWN
    je .wmkeydown
    cmp [wmsg],WM_DESTROY
    je .wmdestroy
  .defwndproc:
    invoke DefWindowProc,[hwnd],[wmsg],[wparam],[lparam]
    jmp .finish
  .wmkeydown:
    cmp [wparam],VK_UP
    jne .vescape

    xor eax,eax
    jmp .finish
  .vescape:

    cmp [wparam],VK_SPACE
    jne @f
    add [speed_flag],1
    and [speed_flag],1
   @@:


    cmp [wparam],VK_ESCAPE
    jne .finish
  .wmdestroy:
    invoke PostQuitMessage,0
    xor eax,eax
  .finish:
    ret
endp

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; g_draw - draw to screen buffer; main program proc
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
proc g_draw

    push ebp
    mov  ebp,esp
    sub  esp,64
    and  ebp,0xfffffff0


    pushad

    xor   eax,eax          ; clear draw buffer
    mov   edi,[screen_ptr]
    movzx ebx,[yres_var]
    movzx ecx,word[xres_var]
    imul  ecx,ebx
    call  fill_aligned

    mov   eax,6000.1
    mov   edi,[Zbuffer_ptr]
    movzx ebx,[yres_var]
    movzx ecx,word[xres_var]
    imul  ecx,ebx
    call  fill_aligned

    fninit
    fild   [ccounter]
    fmul   [one_deg]

    fld    st
    fsincos
    fst   dword[cos_beta]
    fchs
    fstp  [neg_cos_beta]
    fst   [sin_beta]
    fchs
    fstp  [neg_sin_beta]

  @@:
    fld1
    fld1
    faddp
    fmulp

    fsincos
    fst   dword[cos_gamma]
    fchs
    fstp  [neg_cos_gamma]
    fst   [sin_gamma]
    fchs
    fstp  [neg_sin_gamma]

 ; .rot_points
    mov    esi,points_r         ; rotate points
    mov    edi,points_rotated
    movzx  ecx,[points_count_var]
    call   rotor
 ;.rot_normals:
    mov    esi,points_normals
    mov    edi,points_normals_rotated
    movzx  ecx,[points_count_var]
    call   rotor


    cmp   [scale_cnt],912       ; scale counter
    jne   @f
    mov   [scale_dir],1         ; scale direction
  @@:
    cmp   [scale_cnt],0
    jne   @f
    mov   [scale_dir],0
  @@:
    movss xmm0,[scale]
    cmp   [scale_dir],0
    jne   @f
    mulss xmm0,[scale_inc]
    add   [scale_cnt],1
    jmp   .finish_scale
  @@:
    mulss xmm0,[scale_dec]
    sub   [scale_cnt],1
  .finish_scale:
    movss  [scale],xmm0

    invoke CreateThread,NULL,NULL,draw_th0,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID0]
    mov    [hthread0],eax
    invoke  SetThreadPriority,hthread0,THREAD_PRIORITY_TIME_CRITICAL

    invoke CreateThread,NULL,NULL,draw_th1,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID1]
    mov  [hthread1],eax
    invoke  SetThreadPriority,hthread1,THREAD_PRIORITY_TIME_CRITICAL

    invoke SetThreadAffinityMask,[hthread0],1
    invoke SetThreadAffinityMask,[hthread1],2
    invoke ResumeThread,[hthread0]
    invoke ResumeThread,[hthread1]
    invoke WaitForMultipleObjectsEx,2,hthread0,1,-1,0
    invoke CloseHandle,hthread0
    invoke CloseHandle,hthread1

    popad
    add esp,64
    pop ebp
ret

endp

fill_aligned:
;in  eax - dword value to fill,
;    ecx - area size in dwords
;    edi - mem adress
;    addres must be aligned
    mov  ebx,ecx
    and  ebx,111111b  ;
    jz   .fill_sse
  @@:
    mov  [edi],eax
    add  edi,4
    sub  ebx,1
    jnz  @b
  .fill_sse:
    shr    ecx,4
    or     ecx,ecx
    jz     .end
    movd   xmm0,eax
    shufps xmm0,xmm0,0
    movaps xmm1,xmm0
    movaps xmm2,xmm0
    movaps xmm3,xmm0
  @@:
    movaps  [edi],xmm0
    movaps  [edi+16],xmm1
    movaps  [edi+32],xmm2
    movaps  [edi+48],xmm3
    add     edi,64
    sub     ecx,1
    jnz     @b
  .end:
ret

rotor:
 ;  esi -  points/normals to rotate
 ;  edi -  rotated points/normals
 ;  ecx -  points(vertices)/normal vectors count
 .again_rot:
    movlps xmm0,[esi+4]
    shufps xmm0,xmm0,01000100b
    movaps xmm1,[cos_gamma]
    shufps xmm1,xmm1,00110100b
    mulps  xmm0,xmm1
    haddps xmm0,xmm0
    movlps [edi+4],xmm0

    push   dword[esi]
    pop    dword[edi]

    movups xmm2,[edi]
    shufps xmm2,xmm2,10001000b
    movaps xmm3,[cos_beta]
    shufps xmm3,xmm3,00011100b
    mulps  xmm3,xmm2
    haddps xmm3,xmm3

    movss  [edi+8],xmm3
    psrldq xmm3,4
    movss  [edi],xmm3
    add    esi,12
    add    edi,12
    sub    ecx,1
    jnz    .again_rot
ret



init_triangles_normals2:
        fninit
        mov     ebx,triangles_normals
        mov     ebp,triangles
        movzx   ecx,[triangles_count_var]
     @@:
        push    ecx
        push    ebx
        mov     ebx,vectors
        movzx   esi,word[ebp]          ; first point index
        lea     esi,[esi*3]
     ; esi - pointer to 1st 3d point
        shl     esi,2
        add     esi,points_r
        movzx   edi,word[ebp+2]          ; first point index
        lea     edi,[edi*3]
        shl     edi,2
        add     edi,points_r
     ; second point index
     ; edi - pointer to 2nd 3d point

        call    make_vector_r
        add     ebx,12
        mov     esi,edi
        movzx   edi,word[ebp+4]        ; third point index
        lea     edi,[edi*3]
        shl     edi,2
        add     edi,points_r
        call    make_vector_r
        mov     edi,ebx                 ; edi - pointer to 2nd vector
        mov     esi,ebx
        sub     esi,12                  ; esi - pointer to 1st vector
        pop     ebx
        call    cross_product
        mov     edi,ebx
        call    normalize_vector
        add     ebp,6
        add     ebx,12
        pop     ecx
        loop    @b
ret

init_point_normals:
.x equ dword [ebp-4]
.y equ dword [ebp-8]
.z equ dword [ebp-12]
.point_number equ word [ebp-26]
.hit_faces    equ word [ebp-28]

        fninit
        mov       ebp,esp
        sub       esp,28
        mov       edi,points_normals ;[points_normals_ptr]
        mov       .point_number,0
    .ipn_loop:
        mov       .hit_faces,0
        mov       .x,0
        mov       .y,0
        mov       .z,0
        mov       esi,triangles;[triangles_ptr]
        xor       ecx,ecx              ; ecx - triangle number
    .ipn_check_face:
        xor       ebx,ebx              ; ebx - 'position' in one triangle
    .ipn_check_vertex:
        movzx     eax,word[esi+ebx]    ;  eax - point_number
        cmp       ax,.point_number
        jne       .ipn_next_vertex
        push      esi
        mov       esi,ecx
        lea       esi,[esi*3]
        shl       esi,2
        add       esi,triangles_normals

        fld       .x
        fadd      dword[esi]
        fstp      .x
        fld       .y
        fadd      dword[esi+4]
        fstp      .y
        fld       .z
        fadd      dword[esi+8]
        fstp      .z
        pop       esi
        inc       .hit_faces
        jmp       .ipn_next_face
    .ipn_next_vertex:
        add       ebx,2
        cmp       ebx,6
        jne       .ipn_check_vertex
    .ipn_next_face:
        add       esi,6
        inc       ecx
        cmp       cx,[triangles_count_var]
        jne       .ipn_check_face

        fld       .x
        fidiv     .hit_faces
        fstp      dword[edi]
        fld       .y
        fidiv     .hit_faces
        fstp      dword[edi+4]
        fld       .z
        fidiv     .hit_faces
        fstp      dword[edi+8]
        call      normalize_vector
        add       edi,12  ;type vector 3d
        inc       .point_number
        mov       dx,.point_number
        cmp       dx,[points_count_var]
        jne       .ipn_loop

        mov       esp,ebp
ret
norm_aligned_lv:     ; normalise aligned light vectors
;  in edi -> lights_ptr
;     ecx -> lights count
    @@:
     movaps  xmm0,[edi]
     movaps  xmm1,[edi]
     mulps   xmm1,xmm1
     haddps  xmm1,xmm1
     haddps  xmm1,xmm1
     rsqrtps xmm1,xmm1
     mulps   xmm0,xmm1
     movaps  [edi],xmm0
     add     edi,64
     sub     ecx,1
     jnz     @b
ret


include '3DMATH.INC'
include 'procs_a.inc'
include '3bi_fil.inc'
include 'draw.inc'
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; data
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
section '.data' data readable writeable
    _class    TCHAR 'grtest',0
    _title    TCHAR 'Maria Vincit, press space -> speed - full/idle ',0
    _error    TCHAR 'Startup failed.',0
    ccounter    dd  0
    i3          dd  3
    i256        dw  256
    i255d       dd  255
    speed_flag  db  1
    scale_dir   db  1
    scale_cnt   dw  350
    scale_inc   dd  1.0017      ; scale increment
    scale_dec   dd  0.998302885  ; scale decrement

    scale       dd 200.0
    xxadd:
                dw ?
    yyadd       dw ?
    zzadd       dw 1000
;  piD256:
    one_deg     dd  0.01227

align 16

 tex_m2:         dd 124.0, 124.0, 124.0, 124.0
 mask_255f:      dd 255.0, 255.0, 255.0, 255.0
 the_zero:       dd 0, 0, 0, 0
 the_one:        dd 1.0, 1.0, 1.0, 1.0
 mask_h_zero:    dd -1, -1, -1, 0
 abs_val:        dd 0x7fffffff, 0x7fffffff, 0x7fffffff, 0x7fffffff
align 16
 lights_aligned:
    .light_vect   dd   0.0, 0.0, 1.0, 0.0
    .light_c_o:   dd   50.0, 40.0, 88.0, 0.0
    .light_c_min: dd   30.0, 41.0, 30.0, 0.0
    .light_c_max: dd   255.0, 255.0, 5.0, 0.0

       dd   0.5,0.5,5.0, 0.0
       dd   52.0, 47.0, 117.0, 0.0
       dd   1.0,1.0,1.0,0.0
       dd   255.0,2.0,255.0,0.0

       dd   1.0,1.0,1.0,0.0
       dd   163.0,1.0,38.0,0.0
       dd   7.0,3.0,17.0,0.0
       dd   2.0, 255.0, 225.0, 0.0

  lights_aligned_end:

points_count_var    dw 4
triangles_count_var dw 4

points_r:
;tetra_points_r:
dd -1.0, -1.0, 1.0
dd  1.0, -1.0, 1.0
dd  0.0,  1.0, 0.0
dd  0.0, -1.0,-1.0

triangles:
;tetra_triangles:
dw 0,2,1, 3,2,0, 1,2,3,  0,1,3

palette:
include 'palette.inc'

mv16:
   file '128X128X4.hex'

    msg  MSG
    rect RECT
    wc WNDCLASS 0,WindowProc,0,0,NULL,NULL,NULL,1,NULL,_class

    screen_ptr  dd ?
    Zbuffer_ptr dd ?
    xres_var:   dw ?
    yres_var    dw ?

align 16
    cos_beta:
                  rd 1
    sin_beta      dd ?
    neg_cos_beta  dd ?
    neg_sin_beta  dd ?

    cos_gamma:
                  rd 1
    sin_gamma     dd ?
    neg_cos_gamma dd ?
    neg_sin_gamma dd ?

    hwnd      dd ?
    hdc       dd ?
    bmi       BITMAPINFOHEADER

    texmap    rd TEX_X * TEX_Y
    points_rotated:
       rd 20
    triangles_normals:
       rd 20
    points_normals:
       rd 20
    points_normals_rotated:
       rd 20

    vectors     rb 24
    ThreadID0   dd ?
    ThreadID1   dd ?
    ThreadID2   dd ?
    hthread0    rd 1
    hthread1    rd 1

section '.idata' import data readable writeable
    library kernel32,'KERNEL32',\user32,'USER32',\gdi32,'GDI32',\winmm,'WINMM',\
    comdlg32,'COMDLG32'    ;   timing,'timing.dll'
    include '\include\api\kernel32.inc'
    include '\include\api\user32.inc'
    include '\include\api\gdi32.inc'
    include '\include\api\comdlg32.inc'
    import winmm,timeGetTime,'timeGetTime'




